/**
 * Greyd.Blocks Editor Script for core Query Block extension.
 */
( function( wp ) {

	const { createElement: el } = wp.element;
	const { __, _x, sprintf } = wp.i18n;
	const _ = lodash;

	/**
	 * Register custom attributes to core blocks.
	 * - core/query
	 * 
	 * @hook blocks.registerBlockType
	 */
	var registerBlockTypeHook = function(settings, name) {

		if (name == 'core/query') {
			// console.log("register query");
			// keep displayLayout to enable smooth deprecation
			settings.attributes.displayLayout = { type: 'object' };
			// disable core deprecations to enable smooth displayLayout deprecation
			settings.deprecated = [];
			// advanced filter
			settings.attributes.advancedFilter = { type: 'array', default: [] };

			/**
			 * Variations for core/query block.
			 */
			settings.variations.unshift(
				{
					name: "template",
					title: __("Post Slider", 'greyd_hub'),
					description: __("Display posts with Dynamic Templates as sliders, grid, etc.", 'greyd_hub'),
					icon: 'slides',
					category: 'greyd-blocks',
					keywords: [ 'post', 'overview', 'slider', 'query', 'loop' ],
					attributes: {
						query: { perPage: 12, pages: 0, offset: 0, postType: "post", order: "desc", orderBy: "date", author: "", search: "", exclude: [], sticky: "", inherit: false }
					},
					innerBlocks: [
						[ "core/post-template", { layout: { type: 'grid', columnCount: 3, items: 12, responsive: { md: { items: 8, columnCount: 2 }, sm: { items: 4, columnCount: 1 } } }, variation: 'slider' }, [
							[ "greyd/dynamic" ]
						] ]
					],
					example: {
						attributes: {
							query: { perPage: 1, pages: 0, offset: 0, postType: "post", order: "desc", orderBy: "date", author: "", search: "", exclude: [], sticky: "", inherit: false }
						},
						innerBlocks: [
							{
								name: "core/post-template",
								attributes: { layout: { type: 'grid', columnCount: 1, items: 1, variation: 'slider' } },
								innerBlocks: [
									{
										name: "core/group",
										attributes: { layout: { type: 'flex', orientation: 'vertical', justifyContent: 'stretch', verticalAlignment: 'space-between' }, style: { spacing: { margin: { top: "var:preset|spacing|medium", bottom: "var:preset|spacing|medium" } } } },
										innerBlocks: [
											{ name: "core/post-featured-image", attributes: { isLink: true, style: { color: [] } } },
											{ name: "core/post-date", attributes: { style: { typography: { fontSize: "16px", fontWeight: "500" } } } },
											{ name: "core/post-title", attributes: { level: 3, isLink: true, style: { spacing: { margin: { top: "var:preset|spacing|small", bottom: "var:preset|spacing|small" } }, fontSize: "medium" } } },
											{ name: "core/post-excerpt" }
										]
									}
								]
							}
						]
					},
					scope: [ "inserter", "block" ]
				},
				{
					name: "liveSearch",
					title: __("Live Search", 'greyd_hub'),
					description: __("Display posts with live filter capabilities", 'greyd_hub'),
					icon: greyd.tools.getCoreIcon( 'search' ),
					category: 'greyd-blocks',
					keywords: [ 'post', 'search', 'filter', 'slider', 'query', 'loop' ],
					attributes: {
						query: { perPage: 12, pages: 0, offset: 0, postType: "post", order: "desc", orderBy: "date", author: "", search: "", exclude: [], sticky: "", inherit: false }
					},
					innerBlocks: [
						[ "greyd/search", { posttype: "post", greydStyles: { align: "flex-end", flexWrap: "nowrap", responsive: { sm: { wrap: "wrap" } } } }, [
							[ "greyd/search-input", { label: __("Search"), greydStyles: { width: "100%" } } ],
							[ "greyd/search-filter", { parentPosttype: "post", filterBy: "category", label: __("Filter"), placeholder: "select", greydStyles: { width: "max(200px, 50%)" } } ]
						] ],
						[ "core/spacer" ],
						[ "core/post-template", { layout: { type: 'grid', columnCount: 3, items: 12, responsive: { md: { items: 8, columnCount: 2 }, sm: { items: 4, columnCount: 1 } } }, variation: 'slider' }, [
							[ "core/group", { layout: { type: 'flex', orientation: 'vertical', justifyContent: 'stretch', verticalAlignment: 'space-between' }, style: { spacing: { margin: { top: "var:preset|spacing|medium", bottom: "var:preset|spacing|medium" } } } }, [
								[ "core/post-featured-image", { isLink: true, style: { color: [] } } ],
								[ "core/post-date", { style: { typography: { fontSize: "16px", fontWeight: "500" } } } ],
								[ "core/post-title", { level: 3, isLink: true, style: { spacing: { margin: { top: "var:preset|spacing|small", bottom: "var:preset|spacing|small" } } }, fontSize: "medium" } ],
								[ "core/post-excerpt" ]
							] ]
						] ]
					],
					example: {
						attributes: {
							query: { perPage: 1, pages: 1, offset: 0, postType: "post", order: "desc", orderBy: "date", author: "", search: "", exclude: [], sticky: "", inherit: false }
						},
						innerBlocks: [
							{
								name: "greyd/search",
								attributes: { posttype: "post", greydStyles: { align: "flex-end", flexWrap: "nowrap", responsive: { sm: { wrap: "wrap" } } } },
								innerBlocks: [
									{
										name: "greyd/search-input",
										attributes: { label: __("Search"), greydStyles: { width: "100%" } }
									},
									{
										name: "greyd/search-filter",
										attributes: { parentPosttype: "post", filterBy: "category", label: __("Filter"), placeholder: "select", greydStyles: { width: "max(200px, 50%)" } }
									}
								]
							},
							{ name: "core/spacer" },
							{
								name: "core/post-template",
								attributes: { layout: { type: 'grid', columnCount: 1, items: 1, variation: 'slider' } },
								innerBlocks: [
									{
										name: "core/group",
										attributes: { layout: { type: 'flex', orientation: 'vertical', justifyContent: 'stretch', verticalAlignment: 'space-between' }, style: { spacing: { margin: { top: "var:preset|spacing|medium", bottom: "var:preset|spacing|medium" } } } },
										innerBlocks: [
											{ name: "core/post-featured-image", attributes: { isLink: true, style: { color: [] } } },
											{ name: "core/post-date", attributes: { style: { typography: { fontSize: "16px", fontWeight: "500" } } } },
											{ name: "core/post-title", attributes: { level: 3, isLink: true, style: { spacing: { margin: { top: "var:preset|spacing|small", bottom: "var:preset|spacing|small" } }, fontSize: "medium" } } },
											{ name: "core/post-excerpt" }
										]
									}
								]
							}
						]
					},
					scope: [ "inserter", "block" ]
				},
				{
					name: "table",
					title: __("Post Table", 'greyd_hub'),
					description: __("Display posts as table", 'greyd_hub'),
					icon: greyd.tools.getBlockIcon( 'table' ),
					category: 'greyd-blocks',
					keywords: [ 'post', 'table', 'query', 'loop' ],
					attributes: {
						query: { perPage: 10, pages: 0, offset: 0, postType: "post", order: "desc", orderBy: "date", author: "", search: "", exclude: [], sticky: "", inherit: false }
					},
					innerBlocks: [
						[ "greyd/post-table" ]
					],
					example: {
						attributes: {
							query: { perPage: 10, pages: 0, offset: 0, postType: "post", order: "desc", orderBy: "date", author: "", search: "", exclude: [], sticky: "", inherit: false }
						},
						innerBlocks: [
							{ name: "greyd/post-table" }
						]
					},
					scope: [ "inserter", "block" ]
				},
				{
					name: "blank",
					title: __("Blank", 'greyd_hub'),
					icon: 'media-default',
					category: 'greyd-blocks',
					keywords: [ 'post', 'overview', 'query', 'loop' ],
					attributes: {
						query: { perPage: 12, pages: 0, offset: 0, postType: "post", order: "desc", orderBy: "date", author: "", search: "", exclude: [], sticky: "", inherit: false }
					},
					innerBlocks: [
						[ "core/post-template", { layout: { type: 'grid', columnCount: 3 } }, [
							[ "core/paragraph" ]
						] ]
					],
					scope: [ "block" ]
				},
			);
			// console.log(settings);
		}

		return settings;
	};

	wp.hooks.addFilter(
		'blocks.registerBlockType',
		'greyd/hook/query',
		registerBlockTypeHook
	);


	/**
	 * Manipulate attributes before edit.
	 * 
	 * @hook editor.BlockListBlock
	 */
	var editBlockListHook = wp.compose.createHigherOrderComponent( function ( BlockListBlock ) {
		return function ( props ) {
			// console.log(BlockListBlock);
			// console.log(props);

			// 
			// Extensions for inherited query
			// in single, archive and search templates.
			if (
				greyd.data.post_type == "dynamic_template" &&
				props.name == "core/query" &&
				_.has(props.attributes.query, 'inherit') &&
				props.attributes.query.inherit === true
			) {
				// console.log(props);
				// console.log("listing query");

				var query = {
					inherit: true, postType: "post", orderBy: "date", order: "desc", sticky: "", author: "", search: "", exclude: [],
					perPage: null, offset: 0, pages: 0,
				};

				var isSingle = greyd.data.template_type == "single";
				var isArchive = greyd.data.template_type == "archives";
				var isSearch = greyd.data.template_type == "search";
				var postType = greyd.data.post_name.indexOf('-') > -1 ? greyd.data.post_name.split(/-(.*)/s)[1] : null;

				if ( greyd.data.template_type == "woo" ) {
					if ( greyd.data.post_name == "woo-product" ) {
						isSingle = true;
					} 
					else if ( greyd.data.post_name.indexOf("woo-product-") == 0 ) {
						isArchive = true;
					}
				}

				if ( isSingle || isArchive || isSearch ) {
					query['perPage'] = parseInt(greyd.data.posts_per_page);
					if ( postType ) {

						/**
						 * Get the postType slug in archive, single & search templates
						 * based on the curretn template slug.
						 * 
						 * @example
						 * * search-myposttype       -> myposttype
						 * * search-myposttype-news  -> myposttype
						 * * search-my-posttype      -> my-posttype
						 * * search-my-posttype-news -> my-posttype
						 */
						if ( postType.indexOf('-') > -1 ) {
							var postTypeExists = false
							for (const pt of greyd.data.post_types) {
								if (pt.slug == postType) {
									postTypeExists = true;
									break;
								}
							}
							if ( !postTypeExists ) {
								postType = postType.slice( 0, postType.lastIndexOf('-') )
							}
						}

						/**
						 * In FSE templates, the archive template used for categories and tags
						 * is called "archives-category(-general)".
						 */
						if ( postType == 'category' || postType == 'tag' ) {
							postType = 'post';
						}
						// console.log( "dynamic post type:", postType );

						query['postType'] = postType;
					}
				}
				if ( isSingle ) {
					// from vc_post_tp
					query['perPage'] = 1;
				}
				// todo: taxQuery
				if ( !_.isEqual(props.attributes.query, query) ) {
					props.attributes.query = query;
				}
					
				// console.log("inherit query", query);
			}

			return el( BlockListBlock, props );
		};
	}, 'editBlockListHook' );

	wp.hooks.addFilter( 
		'editor.BlockListBlock', 
		'greyd/hook/query/list', 
		editBlockListHook 
	);
	

	/**
	 * Add custom edit controls to core blocks.
	 * - core/query
	 * 
	 * @hook editor.BlockEdit
	 */
	var editBlockHook = wp.compose.createHigherOrderComponent( function( BlockEdit ) {
		
		return function( props ) {	
			
			/**
			 * =================================================================
			 *                          Query extensions
			 * =================================================================
			 */
			
			/**
			 * Extend query block.
			 */
			if (props.name == "core/query") {

				// console.log(props.attributes);
				// console.log("extending query");

				// sync items per page attributes with core/post-template
				const queryPerPage = get(props.attributes.query, 'perPage');
				if ( queryPerPage ) {
					const wrapper = wp.data.select( "core/block-editor" ).getBlocksByClientId( props.clientId )?.[0];
					wrapper?.innerBlocks?.forEach( child => {
						if ( child.name == "core/post-template" ) {
							const itemsPerPage = get(child.attributes.layout, 'items');
							// console.log( itemsPerPage, queryPerPage );
							if ( itemsPerPage != queryPerPage ) {
								// console.log("update post-template child");
								wp.data.dispatch('core/block-editor').updateBlockAttributes( child.clientId, { layout: { 
									...child.attributes.layout, 
									items: parseInt( queryPerPage ) 
								} } );
							}
						}
					} );
				}
				
				// advanced filter controls
				var makeFilter = function() {

					// set advancedFilter attributes
					var setFilter = function(param, value, target) {
						var values = props.attributes.advancedFilter;
						var index = target;
						if (typeof target !== 'number') {
							index = parseInt(target.closest('.components-greyd-controlgroup__item').dataset.index);
						}
						if (!_.has(values, index)) {
							console.warn("advancedFilter index "+index+" not available");
							return;
						}
						// console.log("change advancedFilter "+index+": "+param+": "+value);
						values[index][param] = value;
						if (param == 'name') {
							values[index] = { name: value };
						}
						props.setAttributes( { advancedFilter: [ ...values ] } );
					};

					// controls
					var optionsControl = function(atts) {
						var index = atts.index;
						var advancedFilter = props.attributes.advancedFilter[index];
						var value = !_.isEmpty(advancedFilter[atts.param]) ? advancedFilter[atts.param] : '';
						return el( greyd.components.OptionsControl, {
							label: _.has(atts, 'label') ? atts.label : '',
							help: _.has(atts, 'help') ? atts.help : '',
							value: value,
							onChange: function(value) { 
								setFilter( atts.param, value, index );
							},
							options: atts.options
						} );
					};
					var toggleControl = function(atts) {
						var index = atts.index;
						var advancedFilter = props.attributes.advancedFilter[index];
						var value = _.has(advancedFilter, atts.param) ? advancedFilter[atts.param] : false;
						return el( wp.components.ToggleControl, {
							label: _.has(atts, 'label') ? atts.label : '',
							help: _.has(atts, 'help') ? atts.help : '',
							checked: value,
							onChange: function(value) { 
								setFilter( atts.param, value, index );
							},
						} )
					};
					var radioControl = function(atts) {
						var index = atts.index;
						var advancedFilter = props.attributes.advancedFilter[index];
						var value = !_.isEmpty(advancedFilter[atts.param]) ? advancedFilter[atts.param] : '';
						return el( wp.components.RadioControl, {
							label: _.has(atts, 'label') ? atts.label : '',
							help: _.has(atts, 'help') ? atts.help : '',
							selected: value,
							onChange: function(value) { 
								setFilter( atts.param, value, index );
							},
							options: atts.options
						} );
					};
					var tokenField = function(atts) {
						var index = atts.index;
						var advancedFilter = props.attributes.advancedFilter[index];
						var value = !_.isEmpty(advancedFilter[atts.param]) ? advancedFilter[atts.param] : [];
						var suggestions = atts.suggestions.map(function(item) { return item.label } );
						var ids = atts.suggestions.map(function(item) { return item.value } );
						value = value.map(function(item) { 
							if (ids.indexOf(item) == -1) return item;
							return suggestions[ids.indexOf(item)];
						});
						return el( wp.components.FormTokenField, {
							__experimentalExpandOnFocus: true,
							__experimentalShowHowTo: _.has(atts, 'help') && atts.help,
							value: value,
							suggestions: suggestions,
							onChange: function(value) {
								value = value.map(function(item) { 
									if (suggestions.indexOf(item) == -1) return item;
									return ids[suggestions.indexOf(item)];
								});
								setFilter( atts.param, value, index );
							},
							tokenizeOnBlur: true
						} );
					};
					var DateControl = function(atts) {
						var index = atts.index;
						var advancedFilter = props.attributes.advancedFilter[index];
						var value = !_.isEmpty(advancedFilter[atts.param]) ? advancedFilter[atts.param] : '';
						return el( greyd.components.DatePickerPopupControl, {
							label: _.has(atts, 'label') ? atts.label : null,
							help: _.has(atts, 'help') ? atts.help : null,
							empty: _.has(atts, 'empty') ? atts.empty : null,
							value: value,
							onChange: function(value) { 
								setFilter( atts.param, value, index );
							}
						} );
					};

					var help = function(name, not=false) {
						if (name == 'taxonomy') {
							if (not) return __("Exclude posts with assigned terms from a taxonomy", 'greyd_hub');
							else return __("Show only posts with assigned terms of a taxonomy", 'greyd_hub');
						}
						if (name == 'author') {
							if (not) return __("Exclude posts from specific authors", 'greyd_hub');
							else return __("Show only posts from specific authors", 'greyd_hub');
						}
						if (name == 'include') {
							if (not) return __("Exclude individual posts", 'greyd_hub');
							else return __("Select individual posts to be shown", 'greyd_hub');
						}
						if (name == 'meta') {
							if (not) return __("Exclude posts with meta field property", 'greyd_hub');
							else return __("Show only posts with meta field property", 'greyd_hub');
						}
						if (name == 'order') {
							if (not) return __("Choose the order in which the posts should be sorted in descending order", 'greyd_hub');
							else return __("Choose the order in which the posts should be sorted in ascending order", 'greyd_hub');
						}
						if (name == 'date') {
							if (not) return __("Exclude posts from a specific date time range", 'greyd_hub');
							else return __("Show only posts from a specific date time range", 'greyd_hub');
						}
						return "";
					};
				
					// selected posttype
					var posttype = 'post';
					if ( _.has(props.attributes, 'query') && _.has(props.attributes.query, 'postType') ) {
						posttype = props.attributes.query.postType;
					}

					// get posts of posttype
					var getPosts = function() {
						var posts = {};
						var advancedFilter = [];
						if ( _.has(greyd.data.all_posts, posttype) ) {
							for (var i=0; i<greyd.data.all_posts[posttype].length; i++) {
								var post = greyd.data.all_posts[posttype][i];
								var title = post['id']+": "+post['title'];
								var language = "default";
								if ( _.has(post, 'lang') && !_.isEmpty(post.lang) ) {

									// WPML
									if ( typeof post.lang === 'object' && post.lang.language_code != "" ) {
										// console.log(value.lang);
										language = post.lang.language_code;
										title += " ("+language+")";
										if ( language != greyd.data.language.post.language_code && language != greyd.data.language.default ) {
											advancedFilter.push(post['id']);
										}
										else if ( _.has(post.lang, 'original_id' )) {
											advancedFilter.push(post.lang['original_id']);
										}
									}
									// Polylang
									else if ( typeof post.lang === 'string' && post.lang != "" ) {
										language = post.lang;
										title += " ("+language+")";
										if ( language != greyd.data.language.post && language != greyd.data.language.default ) {
											advancedFilter.push(post['id']);
										}
									}
								}
								if ( !_.has(posts, language) ) posts[language] = [];
								posts[language].push( { value: post['id'], label: title } );
							}
						}
						var final = [];
						for (var [lang, options] of Object.entries(posts)) {
							var advancedFiltered = [];
							if (advancedFilter.length == 0) advancedFiltered = options;
							else {
								for (var i=0; i<options.length; i++) {
									// console.log(options[i]);
									if (advancedFilter.indexOf(options[i].value) == -1)
										advancedFiltered.push(options[i]);
									// else console.log("skipped: "+options[i].value);
								}
							}
							if (advancedFiltered.length > 0) {
								if (lang == 'default') {
									final = [ ...advancedFiltered, ...final ];
								}
								else {
									final = [ ...final, ...advancedFiltered ];
								}
							}
						}
						return final;
					};
					// is posttype hierarchical
					var isHierarchical = function() {
						if ( posttype == 'post' ) return false;
						else if ( posttype == 'page' ) return true;
						else {
							for (var i=0; i<greyd.data.post_types.length; i++) {
								if ( greyd.data.post_types[i]['slug'] == posttype ) {
									if ( _.has(greyd.data.post_types[i]['arguments'], 'hierarchical') ) return true;
									break;
								}
							}
							return false;
						}
					};

					// get taxonomies of posttype
					var getTaxonomies = function() {
						var final = [];
						if ( _.has(greyd.data.all_taxes, posttype) ) {
							for (var i=0; i<greyd.data.all_taxes[posttype].length; i++) {
								var tax = greyd.data.all_taxes[posttype][i];
								if (tax['public']) {
									final.push( { value: tax['slug'], label: tax['title'] } );
								}
							}
						}
						if (final.length == 0) {
							final.push( { value: '', label: sprintf( __("\"%s\" has no taxonomies", 'greyd_hub'), posttype), disabled: true } );
						}
						else {
							final.unshift( { value: '', label: __("Select taxonomy", 'greyd_hub') } );
						}
						return final;
					};
					// is taxonomy hierarchical
					var isHierarchicalTax = function(tax) {
						if ( _.has(greyd.data.all_taxes, posttype ) ) {
							for (var i=0; i<greyd.data.all_taxes[posttype].length; i++) {
								if ( greyd.data.all_taxes[posttype][i]['slug'] == tax ) {
									if (greyd.data.all_taxes[posttype][i]['hierarchical']) return true;
								}
							}
						}
						return false;
					};
					// get terms of taxonomy
					var getTerms = function(tax) {
						var final = [];
						if ( _.has(greyd.data.all_taxes, posttype ) ) {
							for (var i=0; i<greyd.data.all_taxes[posttype].length; i++) {
								if ( greyd.data.all_taxes[posttype][i]['slug'] == tax ) {
									for (var j=0; j<greyd.data.all_taxes[posttype][i]['values'].length; j++) {
										var term = greyd.data.all_taxes[posttype][i]['values'][j];
										var title = term['title']+" ("+term['count']+")";
										final.push( { value: term['id'], label: title } );
									}
									break;
								}
							}
						}
						return final;
					};

					// get all meta fields of posttye
					var getFields = function() {
						var final = [];
						for (var i=0; i<greyd.data.post_types.length; i++) {
							if ( greyd.data.post_types[i]['slug'] == posttype && _.has(greyd.data.post_types[i], 'fields') ) {
								for (var j=0; j<greyd.data.post_types[i]['fields'].length; j++) {
									var field = greyd.data.post_types[i]['fields'][j];
									if (
										!_.has(field, "type") ||
										field["type"] === 'hr' ||
										field["type"] === 'space' ||
										field["type"] === 'headline' ||
										field["type"] === 'descr'
									) continue;
									final.push( { value: field['name'], label: field['label'] } );
								}
								break;
							}
						}
						if (final.length == 0) {
							final.push( { value: '', label: sprintf( __("\"%s\" has no meta fields", 'greyd_hub'), posttype), disabled: true } );
						}
						else {
							final.unshift( { value: '', label: __("Select meta field", 'greyd_hub') } );
						}
						return final;
					};

					// get all authors/users
					var getAuthors = function() {
						var final = [];
						for (var i=0; i<greyd.data.users.length; i++) {
							final.push( { value: greyd.data.users[i]['id'], label: greyd.data.users[i]['display_name'] } );
						}
						return final;
					};
					// has posttype author support
					var hasAuthors = function() {
						if ( posttype == 'post' || posttype == 'page' ) return true;
						else {
							for (var i=0; i<greyd.data.post_types.length; i++) {
								if ( greyd.data.post_types[i]['slug'] == posttype ) {
									if ( _.has(greyd.data.post_types[i]['supports'], 'author') ) return true;
									break;
								}
							}
							return false;
						}
					}

					// is setting postviews_counter enabled
					var postviewsEnabled = function() {
						return _.has(greyd.data.settings, 'advanced_search') && greyd.data.settings.advanced_search.postviews_counter == "true";
					}

					// make controls
					var controls = [];
					if (!_.isEmpty(props.attributes.advancedFilter)) {
						// console.log(props.attributes.advancedFilter);
						var orderOnce = false;
						for (var i=0; i<props.attributes.advancedFilter.length; i++) {
							var advancedFilter = props.attributes.advancedFilter[i];
							if (advancedFilter.name == 'order') orderOnce = true;
							controls.push(
								el( 'div', {
									className: 'components-greyd-controlgroup__item',
									'data-index': i,
								}, [
									// remove
									el( wp.components.Button, { 
										className: "components-greyd-controlgroup__remove",
										onClick: (event) => {
											var index = parseInt(event.target.closest('.components-greyd-controlgroup__item').dataset.index);
											// console.log("remove advancedFilter "+index);
											var values = props.attributes.advancedFilter;
											values.splice(index, 1);
											props.setAttributes( { advancedFilter: [ ...values ] } ); 
										},
										title: __("Remove filter", 'greyd_hub')
									}, el( wp.components.Icon, { icon: 'no-alt' } ) ),

									// advancedFilter
									optionsControl( {
										index: i,
										label: __('Filter', 'greyd_hub'),
										param: 'name',
										options: [
											{ value: '', label: __("Select filter", 'greyd_hub') },
											{ value: 'taxonomy', label: __("Filter by taxonomy", 'greyd_hub') },
											{ value: 'meta', label: __("Filter by meta field", 'greyd_hub') },
											hasAuthors() && { value: 'author', label: __('Filter by authors', 'greyd_hub') },
											{ value: 'date', label: __('Filter by date', 'greyd_hub') },
											{ value: 'include', label: __("Select individual posts", 'greyd_hub') },
											{ value: 'order', label: __("Sort results", 'greyd_hub'), ...orderOnce ? { disabled: true } : {} },
										]
									} ),

									// taxonomy
									advancedFilter.name == 'taxonomy' && [
										optionsControl( {
											index: i,
											label: __("Taxonomy", 'greyd_hub'),
											param: 'taxonomy',
											options: getTaxonomies(),
										} ),
										!_.isEmpty(advancedFilter.taxonomy) && [
											tokenField( {
												index: i,
												param: 'terms',
												suggestions: [
													{ 
														value: 'current_terms', 
														label: advancedFilter.not == true ? 
															__("Exclude terms of the current post", 'greyd_hub') : 
															__("Terms of the current post", 'greyd_hub') 
													},
													{ 
														value: 'current_archive_terms', 
														label: advancedFilter.not == true ? 
															__("Exclude terms of the current archive", 'greyd_hub') : 
															__("Terms of the current archive", 'greyd_hub') 
													},
													{
														value: 'any_terms',
														label: advancedFilter.not == true ? 
															__("Exclude posts with terms", 'greyd_hub') : 
															__("Posts with terms", 'greyd_hub') 
													},
													...getTerms(advancedFilter.taxonomy),
												]
											} ),
											isHierarchicalTax(advancedFilter.taxonomy) && toggleControl( {
												index: i,
												param: 'children',
												label: advancedFilter.not == true ? 
													__("Do not exclude subordinate terms", 'greyd_hub') : 
													__("Exclude subordinate terms", 'greyd_hub')
											} )
										]
									],

									// meta
									advancedFilter.name == 'meta' && [
										optionsControl( {
											index: i,
											param: 'meta',
											options: getFields(),
										} ),
										!_.isEmpty(advancedFilter.meta) && [
											radioControl( {
												index: i,
												param: 'operator',
												options: advancedFilter.not ? [
													{ value: '', label: _x( "Is not set", 'small', 'greyd_hub' ) },
													{ value: 'is', label: _x( "Has not the value:", 'small', 'greyd_hub' ) },
													{ value: 'has', label: _x( "Does not contain the value:", 'small', 'greyd_hub' ) },
												] : [
													{ value: '', label: _x( "Is set", 'small', 'greyd_hub' ) },
													{ value: 'is', label: _x( "Has the value:", 'small', 'greyd_hub' ) },
													{ value: 'has', label: _x( "Contains the value:", 'small', 'greyd_hub' ) },
												]
											} ),
											!_.isEmpty(advancedFilter.operator) && [
												tokenField( {
													index: i,
													param: 'search',
													suggestions: [
														{ 
															value: 'current_meta', 
															label: advancedFilter.not == true ? 
																__("Exclude field value of the current post", 'greyd_hub') : 
																__("Field value of the current post", 'greyd_hub') 
														}
													],
													help: true
												} )
											]
										]
									],

									// author
									advancedFilter.name == 'author' && [
										tokenField( {
											index: i,
											param: 'author',
											suggestions: [
												{ 
													value: 'current_author', 
													label: advancedFilter.not == true ? 
														__("Exclude author of the current post", 'greyd_hub') : 
														__("Author of the current post", 'greyd_hub') 
												},
												{
													value: 'any_author',
													label: advancedFilter.not == true ? 
														__("Exclude posts with author", 'greyd_hub') : 
														__("Posts with author", 'greyd_hub') 
												},
												...getAuthors(),
											]
										} ),
									],

									// date
									advancedFilter.name == 'date' && [
										DateControl( {
											index: i,
											label: __( "Start date", 'greyd_hub' ),
											// empty: __( "no start date set", 'greyd_hub' ),
											empty: advancedFilter.not == true ? 
												__( "Exclude all posts, begining from the first", 'greyd_hub' ) :
												__( "All posts, begining from the first", 'greyd_hub' ),
											param: 'start',
										} ),
										DateControl( {
											index: i,
											label: __( "End date", 'greyd_hub' ),
											// empty: __( "no end date set", 'greyd_hub' ),
											empty: advancedFilter.not == true ? 
												__( "Exclude all posts until now", 'greyd_hub' ) :
												__( "All posts until now", 'greyd_hub' ),
											param: 'end',
										} ),
										!_.isEmpty(advancedFilter.start) && !_.isEmpty(advancedFilter.end) && [
											( (new Date(advancedFilter.start).getTime()) > (new Date(advancedFilter.end).getTime()) ) && [
												el( wp.components.Tip, {}, [
													el( 'div', {
														style: { margin: '10px 0' }
													}, __('The end date should be after the start date.', 'greyd_hub'), )
												] )
											]
										]
									],


									// include
									advancedFilter.name == 'include' && [
										tokenField( {
											index: i,
											param: 'include',
											suggestions: [
												{ 
													value: 'current_post', 
													label: advancedFilter.not == true ? 
														__("Exclude current post", 'greyd_hub') : 
														__("Current post", 'greyd_hub') 
												},
												...getPosts(),
											]
										} ),
										isHierarchical() && toggleControl( {
											index: i,
											param: 'children',
											label: __("Only subordinate posts", 'greyd_hub'),
										} ),
									],

									// order
									advancedFilter.name == 'order' && [
										optionsControl( {
											index: i,
											label: __("Order", 'greyd_hub'),
											param: 'order',
											options: advancedFilter.not ? [
												// absteigend
												{ value: '', label: __("Select sorting", 'greyd_hub') },
												{ value: 'date', label: __("Old to new", 'greyd_hub') },
												{ value: 'title', label: __('Z → A', 'greyd_hub') },
												{ value: 'meta', label: __("Meta descending", 'greyd_hub') },
												isHierarchical() && { value: 'menu_order', label: __("Order descending", 'greyd_hub') },
												postviewsEnabled() && { value: 'views', label: __("Least read", 'greyd_hub') },
												{ value: 'random', label: __("Random", 'greyd_hub') },
											] : [
												// aufsteigend
												{ value: '', label: __("Select sorting", 'greyd_hub') },
												{ value: 'date', label: __("New to old", 'greyd_hub') },
												{ value: 'title', label: __('A → Z', 'greyd_hub') },
												{ value: 'meta', label: __("Meta ascending", 'greyd_hub') },
												isHierarchical() && { value: 'menu_order', label: __("Order ascending", 'greyd_hub') },
												postviewsEnabled() && { value: 'views', label: __("Most read", 'greyd_hub') },
												{ value: 'random', label: __("Random", 'greyd_hub') },
											]
										} ),
										advancedFilter.order == 'meta' && [
											optionsControl( {
												index: i,
												param: 'meta',
												options: getFields(),
											} ),
											!_.isEmpty(advancedFilter.meta) && [
												radioControl( {
													index: i,
													param: 'operator',
													options: [
														{ value: '', label: _x( "Auto", 'small', 'greyd_hub' ) },
														{ value: 'alphabetical', label: _x( "Alphabetical", 'small', 'greyd_hub' ) },
														{ value: 'chronological', label: _x( "Chronological", 'small', 'greyd_hub' ) },
														{ value: 'numeric', label: _x( "Numerical", 'small', 'greyd_hub' ) }
													]
												} ),

											]
										]
									],

									// negate
									advancedFilter.name != '' && [
										toggleControl( {
											index: i,
											param: 'not',
											label: /*advancedFilter.name != 'order' ? __('ausschließen', 'greyd_hub') :*/ __("Reverse / exclude", 'greyd_hub'),
										} ),
										el( 'p', { className: "greyd-inspector-help" }, help(advancedFilter.name, advancedFilter.not) )
									],
								] )
							);
						}
					}
					return controls;
				}

				const isInherit = has(props.attributes, 'query') && has(props.attributes.query, 'inherit') && props.attributes.query.inherit;

				return el( wp.element.Fragment, { }, [
					// original block
					el( BlockEdit, props ),
					// sidebar
					el( wp.blockEditor.InspectorControls, { }, [
						// advancedFilter
						el( greyd.components.AdvancedPanelBody, { 
							title: __("Advanced Filter", 'greyd_hub'),
							initialOpen: !_.isEmpty(props.attributes.advancedFilter),
							holdsChange: !_.isEmpty(props.attributes.advancedFilter)
						}, [
							el( 'div', { className: 'components-greyd-controlgroup'}, [
								makeFilter(),
								el( wp.components.Button, {
									className: 'components-greyd-controlgroup__add'+( _.isEmpty(props.attributes.advancedFilter) ? ' group_is_empty': '' ),
									onClick: function() {
										// console.log('adding parameter');
										var values = !_.isEmpty(props.attributes.advancedFilter) ? props.attributes.advancedFilter : [];
										values.push( { name: '' } );
										props.setAttributes( { advancedFilter: [ ...values ] } ); 
									},
									title: __("Add filter", 'greyd_hub')
								}, [
									el( wp.components.Icon, { icon: 'plus-alt2' } ),
									_.isEmpty(props.attributes.advancedFilter) ? el( 'span', {}, __("Add filter", 'greyd_hub') ) : null
								] )
							] )
						] )
					] ), 
				] );

			}

			// return original block
			return el( BlockEdit, props );
		};

	}, 'editBlockHook' );

	wp.hooks.addFilter( 
		'editor.BlockEdit', 
		'greyd/hook/query/edit', 
		editBlockHook 
	);


} )( 
	window.wp
);