/**
 * Query Block frontend features:
 * Responsive columns, sorting & filtering.
 * 
 * @since 1.5.0
 */
document.addEventListener("DOMContentLoaded", function() {

	var mediaQueries = { trigger: {} };
	function initQueries() {
		var queryBlocks = document.querySelectorAll('.wp-block-query .greyd-posts-slider[live-query="true"]');
		if (queryBlocks.length == 0) return;
		
		// get wp rest api url from localized script
		var requestURL = greyd.rest_api.root + greyd.rest_api.routes.livequery;

		// initial breakpoint to trigger events on load
		let initialBreakpoint = null;

		// live query change function
		const triggerChange = function(obj, perPage = false) {

			// get block data
			var block = JSON.parse(obj.wrap.getAttribute("data-block-data"));

			let taxQuery = {};

			/**
			 * PostTemplate modifications
			 */
			// modify sorting attributes
			if (obj.sorting.length > 0) {
				var sort = obj.sorting[0].selectedOptions[0].value.split(' ');
				block.attrs.query.query.orderBy = sort[0];
				block.attrs.query.query.order = sort[1].toLowerCase();
			}
			// modify filter (taxQuery) attributes
			if (obj.filter.length > 0) {
				obj.filter.forEach(function(filter) {
					const val = filter.value.split(',');
					val.forEach(function(f) {
						if (f != "") {
							if ( f.indexOf('|') > 0 ) {
								var [ slug, id ] = f.split('|');
								if (typeof taxQuery[slug] === 'undefined') taxQuery[slug] = [];
								taxQuery[slug].push(parseInt(id));
							}
							else {
								var slug = filter.name;
								if (typeof taxQuery[slug] === 'undefined') taxQuery[slug] = [];
								taxQuery[slug].push(f);
							}
						}
					})
					// console.log( filter, val, taxQuery );
				});
				block.attrs.query.query.taxQuery = taxQuery;
				block.attrs.query.query.taxQueryRelationship = 'AND';
			}
			// modify perPage attribute
			if (obj.perPage.length > 0) {
				var ppp = obj.perPage[0].selectedOptions[0].value;
				block.attrs.query.query.perPage = ppp;
			} else if (perPage) {
				block.attrs.query.query.perPage = perPage;
			}

			/**
			 * Search modifications
			 */
			// text input
			if (obj.search.input?.length > 0) {
				block.attrs.query.query.search = obj.search.input[0].value;
			}
			// sorting
			if (obj.search.sorting?.length > 0) {
				let sortValue = obj.search.sorting[0].selectedOptions[0].value;
				if (sortValue.indexOf(' ') > 0) {
					var sort = sortValue.split(' ');
					block.attrs.query.query.orderBy = sort[0];
					block.attrs.query.query.order = sort[1].toLowerCase();
				}
			}
			// datepicker
			if (obj.search.date?.length > 0) {
				
				const dateWrapper = obj.search.date[0];
				const filterBy = dateWrapper.getAttribute('data-filter-by');
				const field = dateWrapper.getAttribute('data-field');

				block.attrs.query.query.date = {};
				block.attrs.query.query.date[filterBy] = {};
				block.attrs.query.query.date.filterBy = filterBy;

				const inputs = dateWrapper.parentNode.querySelectorAll(`input[name^="${filterBy}"]`);

				if (filterBy === "post_date" && inputs.length === 2) {
					const after = inputs[0].value;
					const before = inputs[1].value;

					if (after && before) {
						block.attrs.query.query.date[filterBy] = {
							after: after,
							before: before
						};
					}
				} else if (filterBy === "meta_date" || filterBy === "dynamic_meta_date" && inputs.length === 3) {
					let from, to;

					for (const input of inputs) {
						if ( input.matches(`input[name="${filterBy}[from]"`) ) {
							from = input.value;
						} else if ( input.matches(`input[name="${filterBy}[to]"`) ) {
							to = input.value;
						}
					}

					if (from && to) {
						block.attrs.query.query.date[filterBy] = {
							field: field,
							from: from,
							to: to
						};
					}
				}
			}
			// radio buttons
			if (obj.search.buttons?.length > 0) {

				// loop through all buttons (NodeList)
				[ ...obj.search.buttons ].forEach( button => {

					// get taxonomy name
					const taxonomyName = button.getAttribute('name');
					if (typeof taxQuery[taxonomyName] === 'undefined') taxQuery[taxonomyName] = [];

					if ( button.value && button.value.length ) {
						if ( button.value.indexOf(',') > 0 ) {
							const values = button.value.split(',');
							values.forEach(function(v) {
								taxQuery[taxonomyName].push(v);
							});
						} else {
							taxQuery[taxonomyName].push( button.value );
						}
					}
				} );
				
				block.attrs.query.query.taxQuery = taxQuery;
				block.attrs.query.query.taxQueryRelationship = 'AND';
			}


			// abort old request
			// happens when two ore more breakpoints are crossed
			if (obj.request) {
				// console.info("abort "+obj.wrap.id);
				obj.request.abort("abort");
			}

			// make post data
			var postdata = { block: JSON.stringify(block) };
			if ( block.attrs.query.advancedFilter ) {
				// send global post id for advanced filter
				postdata.postId = obj.wrap.getAttribute("data-post-id") ?? -1
			}
			var query_vars = JSON.parse(obj.wrap.getAttribute("data-wp-query"));
			if ( query_vars ) {
				// send wp_query for conditional content
				postdata.wp_query = obj.wrap.getAttribute("data-wp-query");
			}

			// send language
			postdata.lang = greyd.rest_api.lang;

			// send request
			obj.request = jQuery.ajax({
				type: "POST",
				contentType: "application/json; charset=utf-8",
				url: requestURL,
				data: JSON.stringify( postdata ),
			})
			.always( function() {
				obj.spinner[0].style.display = 'none';
				obj.request = false;
			} )
			.done( function onSuccess( response ) {
				// console.log(response);
	
				obj.spinner[0].remove();

				// remove next sibling styleheets
				var styles = obj.wrap.nextElementSibling;
				while (styles && styles.tagName == 'STYLE') {
					styles.remove();
					styles = obj.wrap.nextElementSibling;
				}

				obj.wrap.outerHTML = response.block_content;
				
				// init custom JS events
				// deprecated classic greyd features
				if ( typeof custom_inputs !== 'undefined' && typeof custom_inputs.init === 'function' ) {
					custom_inputs.init();
				}
				if ( typeof posts !== 'undefined' && typeof posts.init === 'function' ) {
					posts.init();
				}
				if ( typeof sorting !== 'undefined' && typeof sorting.init === 'function' ) {
					sorting.init();
				}
				if ( typeof tablesort !== 'undefined' && typeof tablesort.init === 'function' ) {
					tablesort.init();
				}
				// input (multiselects, selects, checkboxes, radios) features FSE
				if ( typeof greyd.input !== 'undefined' && typeof greyd.input.init === 'function' ) {
					greyd.input.init();
				}
				// query (post-slider) features FSE
				if ( typeof greyd.query !== 'undefined' && typeof greyd.query.init === 'function' ) {
					greyd.query.init();
				}
				if ( typeof greyd.trigger !== 'undefined' && typeof greyd.trigger.init === 'function' ) {
					greyd.trigger.init();
				}
				if ( typeof greyd.lazyload !== 'undefined' && typeof greyd.lazyload.init === 'function' ) {
					greyd.lazyload.init();
				}
				if ( typeof greyd.scrollObserver !== 'undefined' && typeof greyd.scrollObserver.init === 'function' ) {
					greyd.scrollObserver.init();
				}
				if ( typeof greyd.animations !== 'undefined' && typeof greyd.animations.init === 'function' ) {
					greyd.animations.init();
				}
				// re-init popovers
				if ( typeof greyd.popover !== 'undefined' && typeof greyd.popover.initPopover === 'function' ) {
					greyd.popover.initPopover();
				}
				// re-init forms
				if ( typeof greyd.forms !== 'undefined' && typeof greyd.forms.initForms === 'function' ) {
					greyd.forms.initForms();
				}
			
				// re-init this
				initQueries();
			} )
			.fail( function onError( XHR, textStatus ) {
				if (textStatus == "abort") {
					// console.info("old call aborted");
				}
				else {
					console.error("Failed to fetch Data: " + textStatus);
					console.log(XHR);
				}
			} );

			// mod html
			// show spinner
			// obj.spinner[0].style = '';
			obj.spinner[0].style.display = 'block';
			obj.spinner[0].style.position = 'absolute';
			// obj.spinner[0].style.top = 0;
			obj.spinner[0].style.marginTop = '-'+obj.wrap.clientHeight+'px';
			// hide greyd-posts-slider
			// obj.wrap.style.display = 'none';
			obj.wrap.style.opacity = 0;

		};

		// media query trigger function
		const triggerMatch = function(event) {
			// console.log(event);
			if (event.matches) {
				// console.log(event);
				var bp = event.target.bp;
				// console.log('Switch to '+bp+' screen.');
				Object.keys(mediaQueries.trigger).forEach(function(key, index) {
					// console.log(key);
					// console.log(mediaQueries.trigger[key]);
					var items = mediaQueries.trigger[key].items;
					if (typeof items[bp] !== 'undefined') {
						// console.info('Switch '+key+' to '+items[bp]+' items per page.');
						triggerChange(mediaQueries.trigger[key].obj, items[bp]);
					}
				});
			} 
		};

		queryBlocks.forEach(function(wrapper, key) {
			// console.log(wrapper);
			let obj = {
				wrap: wrapper,
				sorting: wrapper.querySelectorAll('.sorting select'),
				perPage: wrapper.parentElement.querySelectorAll('.perPage select'),
				filter: wrapper.parentElement.querySelectorAll('.filter input, .filter select'),
				spinner: wrapper.parentElement.querySelectorAll('.loading_spinner_wrapper'),
				search: false,
				request: false
			};
			if ( wrapper.closest('.wp-block-query').querySelector('.wp-block-greyd-search.greyd-search-form') ) {
				var elements = wrapper.closest('.wp-block-query').querySelector('.wp-block-greyd-search.greyd-search-form');
				obj.search = {
					form: elements,
					input: elements.querySelectorAll('input[type=search]'),
					sorting: elements.querySelectorAll('.sorting select'),
					filter: elements.querySelectorAll('.filter input, .filter select'),
					date: elements.querySelectorAll('.greyd-datepicker-input'),
					buttons: elements.querySelectorAll('.wp-block-greyd-search-filter-buttons input[type=hidden]')
				}
			}
			// console.log(obj);

			/**
			 * Add select and multiselect events
			 * need to use jQuery as eventlistener
			 * because multiselect trigger is fired with jQuery
			 */
			jQuery([...obj.filter, ...obj.perPage]).off('change').on('change', function(event) {
				triggerChange(obj);
			});

			/**
			 * Search events
			 */
			if ( obj.search ) {
				const isEnterKey = (e) => {
					// lock the enter key
					var keyCode = e.keyCode || e.which;
					if (keyCode === 13) {
						e.preventDefault();
						return true;
					}
					return false;
				}
				// input
				jQuery(obj.search.input).off("keypress keyup search").on('keypress', function(e) {
					if ( isEnterKey(e) ) return false;
				}).on("keyup search", function(e) {
					if ( isEnterKey(e) ) return false;
					triggerChange(obj);
				});
				// sorting and filter and date
				jQuery([ ...obj.search.sorting, ...obj.search.filter, ...obj.search.date, ...obj.search.buttons ]).off("change").on("change", function(e) {
					triggerChange(obj);
				});
				// submit
				jQuery(obj.search.form).off("submit").on("submit", function(e) {
					e.preventDefault();
					triggerChange(obj);
				});
			}

			/**
			 * Add media query events
			 */
			let perPage = obj.wrap.getAttribute("data-perPage");
			if ( perPage ) {
				
				// get perPage data
				perPage = JSON.parse(perPage);
				// console.log(perPage);
				const bps = [ 'sm', 'md', 'lg', 'xl' ];
				const breakpoints = perPage['breakpoints'];
				const current = perPage['current'];
				let items = {};
				if (typeof perPage['items']['xl'] !== 'undefined')
					items = { 'sm': perPage['items']['xl'], 'md': perPage['items']['xl'], 'lg': perPage['items']['xl'], 'xl': perPage['items']['xl'] };
				if (typeof perPage['items']['lg'] !== 'undefined')
					items = { ...items, 'sm': perPage['items']['lg'], 'md': perPage['items']['lg'], 'lg': perPage['items']['lg'] };
				if (typeof perPage['items']['md'] !== 'undefined')
					items = { ...items, 'sm': perPage['items']['md'], 'md': perPage['items']['md'] };
				if (typeof perPage['items']['sm'] !== 'undefined')
					items = { ...items, 'sm': perPage['items']['sm'] };
				// console.log(items);

				// add to trigger list
				mediaQueries.trigger[obj.wrap.id] = { obj: obj, current: current, items: items };

				/**
				 * make media query conditions once on init of first query
				 */
				bps.forEach(function(bp, index) {
					if (typeof mediaQueries[bp] === 'undefined') {
						const breakpoint = breakpoints[bp];
						const conditions = [];
						if (index == 0) conditions.push('(max-width: '+breakpoint+'px)');
						else {
							var previous = breakpoints[bps[index-1]];
							if (index < 3) {
								conditions.push('(min-width: '+previous+'px)');
								conditions.push('(max-width: '+breakpoint+'px)');
							}
							else conditions.push('(min-width: '+previous+'px)');
						}
						// console.log(conditions);

						// match media change event
						mediaQueries[bp] = {};
						mediaQueries[bp].event = window.matchMedia(conditions.join(' and '));
						mediaQueries[bp].event.onchange = triggerMatch;
						mediaQueries[bp].event.bp = bp;
						// console.log(mediaQueries[bp]);
						if (mediaQueries[bp].event.matches && bp != 'xl') {
							// save breakpoint for initial trigger
							initialBreakpoint = bp;
						} 
					}
				});
			}
		});


		// initial trigger
		if ( initialBreakpoint ) {
			triggerMatch({ matches: true, target: { bp: initialBreakpoint } });
		}

		// console.log("query scripts loaded.");
	}
	initQueries();

});